# feedBackParser.R

# Final pseudocode is:
# if (empty) {
#   show empty text
# } else if (nefarious) {
#   show nefarious text
# } else {
#   noDefaults = TRUE
#   GROUP1()
#   ...
#   GROUPM()
#   if (noDefaults) {
#     show global default text
#   }
# }
#
# Each "GROUP" or "ERRORGROUP" or "NOERRORGROUP" defines a set of non-nested if/elseif's optionally with
#   a final else leading to DEFAULT (starting with "noDefaults = FALSE").  The condition
#   for the ERRORGROUP may be empty "()".
# 
# "GLOBALDEFAULT:" catches if and only if nothing else does
#
# Note that VARS is a list of variables in the user's environment, with starting condition set
#   in the RFeedback Shiny environment, and final condition depending on user R commands.
#
# Predefined variables are ERROR (should not be needed by user), NINPUT, INPUT, NOUTPUT, OUTPUT, 
#   NVARS, VARNAMES, VARS, ERRORCODE, ERRORMESSAGE.  INPUT and OUTPUT are character(0) if empty,
#   so grep() is still suitable.  VAR1 through VAR## (where ## equals min(NVARS, MAXVARS)) are
#   also defined. MAXVARS comes from the definition of the RFeedback environment.  ERRORCODE
#   is the single executable code fragment that caused the error.  ERRORMESSAGE is the error message.
#
# If a GROUP starts with "CALC:", then those variables are defined if and when the main group body
#   is entered (i.e., after error or noError is checked).
# Otherwise, acceptable GROUP() sections headers are logical combinations of:
#   CONDITION() is a generic if.  If condition uses backticks, e.g., `x`, then the
#     condition is met only if the variable(s) exist in the environment
#   INPUTHAS(), INPUTLACKS(), OUTPUTHAS(), OUTPUTLACKS(): grep of "()" condition 
#   ENVEMPTY, ENVNOTEMPTY summarize the environment variable count
#   ENVHAS(), ENVLACKS() look for a variable in the environment
#   VARCLASSIS(,), VARCLASSISNT(,) test if var in environment (name or number) exists and is a
#     certain class
#   ERRORMESSAGEHAS(), ERRORMESSAGELACKS(): grep of () in error message
#   ERRORCODEHAS(), ERRORCODELACKS(): grep of () in code causing error
# Each group can end in a "DEFAULT:".

# Example:
# EMPTY:
# "Please type something!"
# NEFARIOUS:
# "Don't be nefarious"
# GROUP(ENVHAS(x) && ENVHAS(y)):
#   VARCLASSIS(x, numeric) && VARCLASSIS(y, numeric):
#     "Terrific!  You created 'x' and 'y' which are both numeric vectors"
#   VARCLASSIS(x, numeric):
#     paste0("Good!  You created numeric vector 'x', but 'y' is ", class(`y`))
#   VARCLASSIS(y, numeric):
#     paste0("Good!  You created numeric vector 'y', but 'x' is ", class(`x`))
#   DEFAULT:
#     "You created both 'x' and 'y', but neither is a numeric vector."
#     "Check the instuctions and/or solutions and try again."
# 
# GROUP(sum(match(c("x","y"), VARNAMES)) == 1):
#   ENVHAS(x):
#     "You created variable 'x', but not variable 'y'."
#   DEFAULT:
#     "You created variable 'x', but not variable 'y'."
#
# ERRORGROUP:
#  ERRORMESSAGEHAS(unexpected numeric constant)
#  "It appears you incorrectly specified a number."
#  DEFAULT:
#  "Your code caused an error.  The error was caused by you code:"
#  cat(ERRORCODE, sep="\n")
#  "R's error message is:"
#  cat(ERRORMESSAGE, sep="\n")
#  
# GLOBALDEFAULT:
#   "You have not created either 'x' or 'y' as required."


feedbackParser = function(text, lineNumOffset) {
  
  lineNumOffset0 = lineNumOffset - 1
  
  # find global default
  globalDefaultLoc = which(regexpr("^[[:space:]]*GLOBALDEFAULT:", text) != -1)
  if (length(globalDefaultLoc) == 0) {
    globalDefault = '"< Sorry, your input was totally unexpected, and there is no feedback! >"'
    globalDefaultLoc = length(text) + 1
  } else if (length(globalDefaultLoc) > 1) {
    stop("cannot have multiple global defaults in RFeedback environment @ ", paste(lineNumOffset0+globalDefaultLoc, collapse=", "))
  } else {
    globalDefault = str_trim(text[(globalDefaultLoc+1) : length(text)])
  }
  #
  emptyLoc = which(regexpr("^[[:space:]]*EMPTY[[:space:]]*:$", text) != -1)
  if (length(emptyLoc) == 0) {
    empty = '"(no input)"'
    emptyLoc = 0
  } else if (length(emptyLoc) > 1) {
    stop("cannot have multiple 'EMPTY's in RFeedback environment @ ", paste(lineNumOffset0+emptyLoc, collapse=", "))
  } else {
    empty = NULL # indicates delayed evaluation
  }
  #
  nefariousLoc = which(regexpr("^[[:space:]]*NEFARIOUS[[:space:]]*:$", text) != -1)
  if (length(nefariousLoc) == 0) {
    nefarious = '"Please do not attempt to use dangerous functions!"'
    nefariousLoc = 0
  } else if (length(nefariousLoc) > 1) {
    stop("cannot have multiple 'NEFARIOUS's in RFeedback environment @ ", paste(lineNumOffset0+nefariousLoc, collapse=", "))
  } else {
    nefarious = NULL # indicates delayed evaluation
  }
  
  # find group starts and capture any empty/nefarious information
  groupStartLoc = which(regexpr("^[[:space:]]*(ERROR|NOERROR)?GROUP", text) != -1)
  nGroups = length(groupStartLoc)
  if (nGroups == 0) stop("RFeedback environments must have at least one [[NO]ERROR]GROUP @ ", lineNumOffset0)
  if (any(groupStartLoc > globalDefaultLoc)) stop("'GLOBALDEFAULT:' must be last @ ", lineNumOffset0+globalDefaultLoc)
  if (any(groupStartLoc < emptyLoc)) stop("'EMPTY:' must be before the first [[NO]ERROR]GROUP @ ", lineNumOffset0+emptyLoc)
  if (any(groupStartLoc < nefariousLoc)) stop("'NEFARIOUS:' must be before the first [[NO]ERROR]GROUP @ ", lineNumOffset0+nefariousLoc)
  if (is.null(empty) && is.null(nefarious)) {
    if (emptyLoc < nefariousLoc) {
      if (emptyLoc+1 == nefariousLoc) stop("'EMPTY' cannot be empty @ ", lineNumOffset0+emptyLoc)
      empty = text[(emptyLoc+1) : (nefariousLoc-1)]
      if (nefariousLoc+1 == groupStartLoc[1]) stop("'NEFARIOUS' cannot be empty @ ", lineNumOffset0+nefariousLoc)
      nefarious = text[(nefariousLoc+1) : (groupStartLoc[1]-1)]
    } else { # nefarious is first
      if (nefariousLoc+1 == emptyLoc) stop("'NEFARIOUS' cannot be empty @ ", lineNumOffset0+nefariousLoc)
      nefarious = text[(nefariousLoc+1) : (emptyLoc-1)]
      if (emptyLoc+1 == groupStartLoc[1]) stop("'EMPTY' cannot be empty @ ", lineNumOffset0+emptyLoc)
      empty = text[(emptyLoc+1) : (groupStartLoc[1]-1)]
    }
  } else if (is.null(empty)) {
    if (emptyLoc+1 == groupStartLoc[1]) stop("'EMPTY' cannot be empty @ ", lineNumOffset0+emptyLoc)
    empty = text[(emptyLoc+1) : (groupStartLoc[1]-1)]
  } else if (is.null(nefarious)) {
    if (nefariousLoc+1 == groupStartLoc[1]) stop("'NEFARIOUS' cannot be empty @ ", lineNumOffset0+nefariousLoc)
    nefarious = text[(nefariousLoc+1) : (groupStartLoc[1]-1)]
  }
  
  # Compute group info
  groupInfo = data.frame(groupStart=groupStartLoc, headerLines=NA_integer_, groupEnd=NA_integer_, 
                         condition=NA_character_, isErrorGroup=NA, isNoErrorGroup=NA)
  groupInfo$groupEnd = c(groupStartLoc[-1] - 1, globalDefaultLoc - 1)
  termColons = which(regexpr(":$", text) != -1)
  okColons = sapply(1:nGroups, 
                    function(n) termColons[termColons>=groupInfo[n,"groupStart"] & termColons<=groupInfo[n,"groupEnd"]][1])
  Sel = is.na(okColons)
  if (sum(Sel)>0) stop("missing ':' for [[NO]ERROR]GROUP @ ", paste(lineNumOffset0+groupInfo$groupStart[Sel], collapse=", "))
  groupInfo$headerLines = okColons - groupInfo$groupStart + 1
  temp = sapply(1:nGroups, 
                function(n) {
                  paste(text[groupInfo[n,"groupStart"] :
                             (groupInfo[n,"groupStart"] + groupInfo[n,"headerLines"] - 1)],
                        collapse=" ")
                })
  groupInfo$isErrorGroup = sapply(temp, function(txt) length(grep("^[[:space:]]*ERRORGROUP", txt)) > 0)
  groupInfo$isNoErrorGroup = sapply(temp, function(txt) length(grep("^[[:space:]]*NOERRORGROUP", txt)) > 0)
  groupInfo$condition = sapply(temp, 
                               function(txt) {
                                 loc = regexpr("[(].*)", txt)
                                 if (loc == -1) return(NA)
                                 return(substring(txt, loc+1, loc+attr(loc,"match.length")-2))
                               })
  Sel = is.na(groupInfo$condition)
  #if (any(Sel)) stop("missing parenthesized condition @ ", paste(lineNumOffset0+groupInfo$groupStart[Sel], collapse=", "))
  if (any(Sel)) groupInfo$condition[Sel] = "TRUE"
  #
  # Allow pre-defined conditions inside [[NO]ERROR]GROUP conditions
  temp = try(sapply(groupInfo$condition, makeCondition))
  Sel = sapply(temp, is, class2="try-error")
  if (any(Sel)) {
    stop(attr(temp[[which(Sel)[1]]], "message"), " @ ", paste(lineNumOffset0+groupInfo$groupStart[Sel], collapse=", "))
  }
  groupInfo$condition = temp
  #
  Sel = with(groupInfo, groupEnd < groupStart + headerLines)
  if (any(Sel)) stop("empty [[NO]ERROR]GROUP: @ ", paste(lineNumOffset0+groupInfo$groupStart[Sel], collapse=", "))
  groupText = lapply(1:nGroups, 
                     function(n) {
                       text[(groupInfo[n,"groupStart"] + groupInfo[n,"headerLines"]) :
                            groupInfo[n,"groupEnd"]]
                     })
  
  # Parse text for each group
  groupDetail = vector("list", nGroups)
  validSections = c("CALC", "CONDITION", "INPUTHAS", "INPUTLACKS", "OUTPUTHAS", "OUTPUTLACKS",
                    "ENVEMPTY", "ENVNOTEMPTY", "ENVHAS", "ENVLACKS",
                    "VARCLASSIS", "VARCLASSISNT", 
                    "ERRORMESSAGEHAS", "ERRORMESSAGELACKS", "ERRORCODEHAS", "ERRORCODELACKS",
                    "DEFAULT")
  reStarOK = c("INPUTHAS", "INPUTLACKS", "OUTPUTHAS", "OUTPUTLACKS","ERRORMESSAGEHAS", 
                "ERRORMESSAGELACKS", "ERRORCODEHAS", "ERRORCODELACKS")
  reStarBad = setdiff(validSections, reStarOK)

  addCat = function(fbText) {
    for (fbLine in 1:length(fbText)) {
      c1 = substring(fbText[fbLine], 1, 1)
      nc = nchar(fbText[fbLine])
      if (c1 %in% c('"', "'") && substring(fbText[fbLine], nc)==c1) {
        fbText[fbLine] = paste0('cat(', substring(fbText[fbLine], 1, nc-1), '\\n', c1, ')')
      }
    }
    fbText = fbText[nchar(fbText)>0]
    return(fbText)
  }

  for (grp in 1:nGroups) {
    # Find block markers in recipe
    grpOffset0 = lineNumOffset0 + groupInfo[grp, "groupStart"]
    recipeOffset0 = grpOffset0 + groupInfo[grp, "headerLines"] - 1
    recipe = str_trim(text[(groupInfo[grp, "groupStart"]+groupInfo[grp, "headerLines"]) : 
                            groupInfo[grp, "groupEnd"]])

    sectionStart = grep("^[!(]*[A-Z]+[*]?[[:space:]]*[:(&|)]", recipe)
    # A single un-named section is taken to be a DEFAULT section
    if (length(sectionStart) > 0) {
      nSections = length(sectionStart)
      sectionContinue = c(FALSE, 
                          regexpr(":$", recipe[sectionStart[-nSections]]) == -1  &
                          sectionStart[2:nSections] == 1 + sectionStart[1:(nSections-1)])
      if (any(sectionContinue)) {
        sectionStart = sectionStart[!sectionContinue]
        nSections = length(sectionStart)
      }
      termColons = which(regexpr(":$", recipe) != -1)
      gD = data.frame(sectionStart=sectionStart, sectionEnd=c(sectionStart[-1]-1, length(recipe)))
      okColons = sapply(1:nSections, 
                        function(n) termColons[termColons>=gD[n,"sectionStart"] & termColons<=gD[n,"sectionEnd"]][1])
      Sel = is.na(okColons)
      if (sum(Sel)>0) stop("missing ':' for feedback section @ ", paste(recipeOffset0+gD$sectionStart[Sel], collapse=", "))
      gD$headerLines = okColons - gD$sectionStart + 1
    } else {
      if (length(recipe)==0) stop("empty GROUP @ ", paste(recipeOffset0+1, collapse=", "))
      nSections = 1
      gD = data.frame(sectionStart=1, sectionEnd=length(recipe))
      gD$headerLines = 0
    }
    gD$condition = sapply(1:nSections,
                          function(sect) {
                            if (nrow(gD)==1 && gD$headerLines==0) return("DEFAULT")
                            txt = paste(recipe[gD$sectionStart[sect] : (gD$sectionStart[sect]+gD$headerLines[sect]-1)], collapse=" ")
                            txt = substring(txt, 1, nchar(txt)-1) # remove ":"
                            chk = gregexpr("[[:space:]&|!][A-Z]+[*]?[[:space:]:&|(]", paste0(" ", txt, " "))[[1]]
                            lenM3 = attr(chk, "match.length") - 3
                            chk = sapply(seq(along.with=lenM3), function(i) substring(txt, chk[i], chk[i]+lenM3[i]))
                            Sel = sapply(sub("*", "", chk, fixed=TRUE), function(x) !x %in% validSections)
                            if (any(Sel)) {
                              stop("invalid feedback section name(s) ", paste0("'", chk[Sel], "'", collapse=", "),
                                   " @ ", recipeOffset0+gD[sect, "sectionStart"])
                            }
                            return(txt)
                          })

    Sel = grep(paste0('(', paste0(reStarBad, '[*]', collapse="|"), ')'), gD$condition)
    if (length(Sel) > 0) {
      stop("no-regexpr star illegal for ", paste(reStarBad, collapse=", ")," @ ", 
           paste0(recipeOffset0+gD[Sel, "sectionStart"], collapse=", "))
    }
    Sel = which(gD$condition == "CALC")
    if (length(Sel)>1) stop("multiple 'CALC's in [[NO]ERROR]]GROUP @ ", paste0(recipeOffset0+gD[Sel, "sectionStart"], collapse=", "))
    if (length(Sel)==1 && Sel!=1) stop("'CALC' must be first in its [[NO]ERROR]]GROUP @ ", paste0(recipeOffset0+gD[Sel, "sectionStart"], collapse=", "))
    if (length(Sel)==1 && nrow(gD)==1) stop("'CALC' cannot be the only section in [[NO]ERROR]]GROUP @ ", paste0(recipeOffset0+gD[Sel, "sectionStart"], collapse=", "))
    Sel = which(gD$condition == "DEFAULT")
    if (length(Sel)>1) stop("multiple 'DEFAULT's in [[NO]ERROR]]GROUP @ ", paste0(recipeOffset0+gD[Sel, "sectionStart"], collapse=", "))
    if (length(Sel)==1 && Sel != nSections) stop("'DEFAULT's must be last in [[NO]ERROR]GROUP @ ", recipeOffset0+gD[Sel, "sectionStart"])
    if (nrow(gD)==1 && gD$headerLines==0) {
      feedbackStart = 1
    } else {
      Sel = ( gD$sectionEnd == (gD$sectionStart + gD$headerLines - 1) )
      if (any(Sel)) stop("empty feedback section(s) @ ", paste0(recipeOffset0+gD[Sel, "sectionStart"], collapse=", "))
      feedbackStart = gD$sectionStart + gD$headerLines
    }
    
    groupDetail[[grp]] = lapply(1:nSections, function(sect) list(condition = gD[sect, "condition"], 
                                                                 line = gD[sect, "sectionStart"],
                                                                 feedback = addCat(recipe[feedbackStart[sect]:gD[sect, "sectionEnd"]])))
  }

  return(list(groupInfo=groupInfo, groupDetail=groupDetail, 
              empty=addCat(empty), nefarious=addCat(nefarious), globalDefault=addCat(globalDefault)))
}


if (exists("testingFeedbackParser")) {
  test = readLines("example.txt")
  jnk = feedbackParser(test, 1, NULL)
  print(jnk)
}
